#!/usr/bin/env python3
import sys, string, random, subprocess, yaml, shutil, requests, time, re
from pathlib import Path
from contextlib import contextmanager
from itertools import islice

MAIN = Path(__file__)
MAIN_DIR = MAIN.parent


class Util:

  def random_string(n=16):
    return ''.join(random.choice(string.ascii_lowercase) for i in range(n))

  @contextmanager
  def tmp_file(content):
    tmp = Path(Util.random_string())
    # print("-- temp file --\n" + content + "\n-- end --")
    try:
      with open(tmp, mode="wt", encoding="utf-8") as file:
        file.write(content)
      yield tmp
    finally:
      tmp.unlink()

  @contextmanager
  def tmp_yaml(obj):
    with Util.tmp_file(yaml.safe_dump(obj, sort_keys=False)) as tmp:
      yield tmp


class Jobs:
  pending = []

  def add(args):
    Jobs.pending.append(subprocess.Popen(args))

  def wait_all():
    for job in Jobs.pending:
      job.communicate()
    Jobs.pending = []


class Pact:
  bin = "pact"
  sources = MAIN_DIR / "pact"
  logs = MAIN_DIR / "log"
  local_port = 8080
  local_host = "http://localhost:{}".format(local_port)
  config = {
    "port": local_port,
    "logDir": str(logs),
    "persistDir": str(logs),
    "pragmas": [],
    "verbose": True
  }

  def start(clear = False):
    if Pact.logs.is_dir() and clear:
      shutil.rmtree(Pact.logs)
    if not Pact.logs.is_dir():
      Pact.logs.mkdir()

    with Util.tmp_yaml(Pact.config) as path:
      Jobs.add([Pact.bin, "-s", str(path)])
      time.sleep(1)

  def send(obj, host = local_host):
    with Util.tmp_yaml(obj) as path:
      tx_json = subprocess.check_output([Pact.bin, "-a", str(path)])
    requests.post(
      "{}/api/v1/send".format(host),
      headers = { "Content-Type": "application/json" },
      data = tx_json
    )

  def resolve(path):
    return str(Pact.sources / path)

  def dependencies(path):
    with open(Pact.sources / path) as file:
      return re.findall(r'\(load "(.*)"\)', file.read())

  def write_macro(path, vars):
    with open(path) as file:
      lines = file.readlines()
    def mk_block(tokens, loc, line):
      negate = False
      if tokens[0] == "not":
        negate = True
        tokens = tokens[1:]
      var = tokens[0]
      if var in vars:
        o = { "active": True, "value": vars[var] ^ negate }
      else:
        o = { "active": False, "value": False }
      return { "loc": loc, "line": line } | o
    stack = []
    edits = False
    for i, line in enumerate(lines):
      if line.startswith("; #"):
        tokens = list(islice(filter(lambda x: len(x) > 0, line[1:-1].split(" ")), 3))
        if tokens[0] == "#if":
          stack.append(mk_block(tokens[1:], i, line))
        elif tokens[0] == "#else":
          stack[-1]["value"] = not stack[-1]["value"]
        elif tokens[0] == "#endif":
          stack.pop()
      elif stack and any(o["active"] for o in stack):
        value = True
        [ value := value & (o["value"] | (not o["active"])) for o in stack ]
        comment = line.startswith(";")
        edits = edits | (value == comment)
        if value and comment:
          off = 2 if line[1] == " " else 1
          lines[i] = line[off:]
        elif not value and not comment:
          lines[i] = "; " + line
    if stack:
      sys.exit("macro parse error in: " + str(path) + "\nstack = " + str(stack))
    elif edits:
      with open(path, "w") as file:
        file.write("".join(lines))

PROJECT_FILES = ['domain-v0.pact', 'util.pact', 'flow.pact', 'poll.pact', 'reputation.pact', 'awa.pact', 'deploy-test.pact']

def help():
  print(WELCOME)
  for k, o in COMMANDS.items():
    print("\n\033[1m" + k + "\033[0m\n    " + o["desc"])
  print()

def serve():
  Pact.start("--restore" not in sys.argv)
  keypairs = [{
    "public": "25eeee5ac9cd8394bcf138587c7fcab79367c08b4adb6a4cc2979b6265b13de0",
    "secret": "326737a0db52dc1747e2e154406997ed76ce7914471bd60dc643f70678a97da4"
  }]
  namespace = "free"
  obj = {
    "data": {
      "bunny": [ keypairs[0]["public"] ],
      "ks": [ "cdc8786723c057845f4330d3a2afe32418cfb971f68d617a38521ffcd5fc49a5" ],
      "ns": namespace
    },
    "keyPairs": keypairs
  }
  def load(path):
    Pact.send({ "codeFile": Pact.resolve(path)} | obj)

  load("root/fungible-v2.pact")
  load("root/fungible-xchain-v1.pact")
  load("root/coin-v5.pact")
  Pact.send({ "code": """
    (module a G (defcap G () true) (defun always () true))
    (define-namespace "{ns}" (create-user-guard (always)) (create-user-guard (always)))
  """.format(ns = namespace)} | obj)
  for f in PROJECT_FILES:
    load(f)

def macro():
  p = input("action may update all pact files in project and conflict with unsaved changes\ncontinue (y/N)? ")
  if p == "y":
    itemize = lambda xs: (xs[0], xs[1] != "0")
    vars = dict([ itemize(a.split("=")) for a in sys.argv[2:] ])
    for f in PROJECT_FILES:
      Pact.write_macro(Pact.resolve(f), vars)
  else:
    print("skipped")

WELCOME = """\033[38;5;15;48;5;212m awa \033[0m
high level utility script. commands:"""

COMMANDS = {
  "serve": {
    "desc": "start local pact server and deploy project modules",
    "exec": serve
  },
  "deps": {
    "desc": "parse and display repl dependencies `(load _)`",
    "exec": lambda: print(Pact.dependencies(sys.argv[2]))
  },
  "macro": {
    "desc": "switch comment blocks within pact code\n    e.g. \033[3m./" + MAIN.name + " macro FAKECHAIN=1\033[0m",
    "exec": macro
  },
  "help": {
    "desc": "display this thing",
    "exec": help
  }
}

if __name__ == "__main__":
  random.seed()

  cmd = sys.argv[1] if len(sys.argv) > 1 and sys.argv[1] in COMMANDS else "help"
  COMMANDS[cmd]["exec"]()

  Jobs.wait_all()
