(namespace (read-msg 'ns))
(define-keyset 'bunny)

(interface awa-persistent-v0

  (defconst GENESIS (util.now)))


(module awa GOVERNANCE

  (use util)
  (use awa-persistent-v0)
  (use reputation-policy-v0 [ outcome ])

  (defconst SELF awa)
  (defconst RESOURCE "OWO")
  (defconst MAX_ACTIVITY 20000.0 @doc "determines stake amount")
  (defconst KDA_STAKE_TARGET 5.0 @doc "based on faucet drops of 20 tokens")
  (defconst STAKE_DURATION (days 7.0))
  (defconst MAX_NAME_LENGTH 64)
  (defconst MAX_DATA_LENGTH 256)
  (defconst RATINGS [ "new" "hot" "top" ])
  (defconst DELETED "" @doc "special data value of deleted revisions")

  ; reserved accounts
  (defconst TREASURY (format "treasury@{}" [ SELF ]))
  (defconst BURNS (create-principal (never)))

  (defcap INTERNAL () true)

  (defcap GOVERNANCE ()
    (compose-capability (INTERNAL))
    (enforce-keyset 'bunny))

  (defcap DOCUMENT (name:string account:string)
    @managed
    (compose-capability (INTERNAL))
    (validate-name name)
    (enforce-account account))

  (defcap REVISION (document:string account:string)
    @managed
    (compose-capability (INTERNAL))
    (enforce-account account))

  (defcap VOTE (document:string ordinal:integer account:string)
    @managed
    (enforce-account account))



  (defschema score
    @model [ (invariant (>= revision 0)) ]
    revision:integer
    fitness:decimal )

  (defschema document
    revisions:integer
    candidates:[object{score}]
    t:time )

  (defschema revision
    @model [ (invariant (>= ordinal 0)) ]
    ordinal:integer
    data:string )

  (deftable documents:{document})
  (deftable revisions:{revision})



  (defun initialize ()
    (with-capability (OPERATE reputation)
      (reputation.create-resource RESOURCE SELF SELF))
    (let ((genesis (at 'genesis (reputation.get-resource RESOURCE))))
      (enforce (= genesis GENESIS) "error: unexpected block time"))

    (create-account TREASURY (get-governance-guard))
    (if (not (has-account BURNS)) (create-account BURNS (never)) ""))

  (defun deposit:decimal (account:string input:decimal)
    (enforce (> input 0.0) "invalid amount")
    (coin.transfer account TREASURY input)
    (let ((output (round (/ input (get-seed-price)) (precision))))
      (with-capability (OPERATE reputation)
        (reputation.seed RESOURCE account output))
      output))

  (defun create-document (name:string account:string data:string)
    (with-capability (DOCUMENT name account)
      (insert documents name
        { 'revisions: 0
        , 'candidates: []
        , 't: (now) })
      (emit-event (REVISION name account))
      (make-revision name account data)))

  (defun create-revision (document:string account:string data:string)
    (with-capability (REVISION document account)
      (make-revision document account data)))

  (defun vote:decimal (document:string ordinal:integer account:string ds:decimal)
    (with-capability (VOTE document ordinal account)
      (let*
        ( (revision (get-revision-id document ordinal))
          (mark
            (with-capability (OPERATE reputation)
              (reputation.vote RESOURCE revision account ds))) )

        (with-read revisions revision { 'data := data }
          (enforce (!= data DELETED) "revision has been deleted"))

        (with-read documents document { 'candidates := candidates }
          (update documents document
            { 'candidates: (update-candidates ordinal candidates mark) }))

        (at 'cost mark))))



  (defun has-account:bool (account:string)
    (>= (try -1.0 (coin.get-balance account)) 0.0))

  (defun enforce-account (account:string)
    (enforce-guard (at 'guard (coin.details account))))

  (defun enforce-governance:bool ()
    (require-capability (GOVERNANCE)))

  (defun get-governance-guard:guard ()
    (create-user-guard (enforce-governance)))

  (defun validate-name:bool (name:string)
    (enforce (not (contains ";" name)) "prohibited character: ;")
    (enforce (not (contains "/" name)) "prohibited character: /")
    (enforce (between 1 MAX_NAME_LENGTH (length name)) "invalid name length"))

  (defun get-revision-id:string (document:string ordinal:integer)
    (format "{}:{}" [ document (int-to-str 16 ordinal) ]))

  (defun rate:[decimal] (mark:object{reputation.mark})
    ; values will be rounded to precision
    (let
      ( (dt (diff-time (now) (at 't (at 'market (at 'swap mark)))))
        (day (days 1.0)) )
      [ ; new
        (/ 1.0 (+ 1.0 (/ dt day)))

        ; hot
        ; (TODO: can be manipulated, use a time-weighted average or something)
        (if (>= dt day) (at 'weight (at 'outflow (at 'area mark))) 0.0)

        ; top
        ; (probably use time-adjusted values that emphasize recent revisions in sub-linear supply emission)
        (at 'flow (at 'outflow (at 'area mark))) ]))

  (defun select-candidate:object{score} (old:object{score} new:object{score})
    (if
      (or
        (= (at 'revision old) (at 'revision new)) ; always keep updated value
        (>= (at 'fitness new) (at 'fitness old)))
      new old))

  (defun update-candidates:[object{score}]
    ( ordinal:integer
      previous:[object{score}]
      mark:object{reputation.mark} )
    (let
      ( (len (length previous))
        (scores
          (map
            (lambda (x) { 'revision: ordinal, 'fitness: (round x (precision)) })
            (rate mark))) )
      (enforce (>= (length scores) len) "error: unexpected length")
      (+
        (zip (select-candidate) previous (take len scores))
        (drop len scores))))

  (defun total-supply:decimal ()
    (let ((supply (- (reputation.get-supply RESOURCE) (get-balance BURNS))))
      (enforce (>= supply 0.0) "error: unexpected value")
      supply))

  (defun get-document:object{document} (document:string)
    (read documents document))

  (defun get-revision:object{document} (revision:string)
    (read documents revision))

  (defun list-documents:[string] ()
    (keys documents))

  (defun get-stake:object{stake} ()
    { 'amount: (/ (at 'source (reputation.get-snap SELF GENESIS)) MAX_ACTIVITY)
    , 'duration: STAKE_DURATION })

  (defun get-seed-price:decimal ()
    (/ KDA_STAKE_TARGET (at 'amount (get-stake))))


  (defun make-revision (document:string account:string data:string)
    (require-capability (INTERNAL))
    (enforce (between 1 MAX_DATA_LENGTH (length data)) "invalid data length")
    (with-read documents document { 'revisions := ordinal, 'candidates := candidates }
      (let*
        ( (id (get-revision-id document ordinal))
          (mark
            (with-capability (OPERATE reputation)
              (at 0 (reputation.create-submission RESOURCE id account (get-stake))))) )
        (insert revisions id { 'ordinal: ordinal, 'data: data })
        (update documents document
          { 'revisions: (+ 1 ordinal)
          , 'candidates: (update-candidates ordinal candidates mark) }))))



  (implements domain-v0)

  (defcap OPERATE:bool (target:module{domain-v0}) true)


  (implements reputation-policy-v0)

  (defun parameters:object{elastic.parameters} ()
    { 'p: 4.0, 'q: 0.5, 'precision: (precision) })

  (defun learning-curve:decimal (t:decimal)
    ; linear inflation (~31,557,600 tokens each year)
    t)

  (defun seeding-curve:decimal (t:decimal)
    ; flat initial mint
    1000000.0)

  (defun density:decimal (t:decimal)
    ; 3 months half-life for unstaked tokens
    ; about equivalent to balances being multiplied by 0.999997364261 each block
    (^ 2.0 (/ (- t) (months 3.0))))

  (defun reflow:decimal (t:decimal outcome:object{outcome})
    ; this function should be defined according to sybil resistance mechanism in use (burning, staking, or DID),
    ; and desired dynamics between reputation specialization and inclusion of new actors
    ; the economical difficulty to increase shares on a given outcome depends on p, q, and shares on other outcomes
    ; caution: values are not continuously updated (on each trade), high rate of change can potentially be abused

    ; early square root map helps supply decentralization,
    ; while later quadractic map reduces rewards on low quality content (i.e. ai spam)
    (^ (at 'shares outcome) (min 2.0 (+ 0.5 (/ t (years 1.0))))))


  (implements fungible-v2)

  (defcap TRANSFER:bool (sender:string receiver:string amount:decimal)
    @managed amount TRANSFER-mgr
    (compose-capability (INTERNAL))
    (compose-capability (OPERATE reputation))
    (enforce (!= sender receiver) "same sender and receiver")
    (enforce-unit amount)
    (enforce (> amount 0.0) "positive amount")
    (enforce-account sender))

  (defun TRANSFER-mgr:decimal (managed:decimal requested:decimal)
    (linear-transfer-manager managed requested))

  (defun transfer:string (sender:string receiver:string amount:decimal)
    (with-capability (TRANSFER sender receiver amount)
      (reputation.transfer RESOURCE sender receiver amount)))

  (defun transfer-create:string
    ( sender:string
      receiver:string
      receiver-guard:guard
      amount:decimal )
    (if (has-account receiver)
      (enforce
        (= receiver-guard (at 'guard (coin.details receiver)))
        "guard mismatch")
      (create-account receiver receiver-guard))
    (transfer sender receiver amount))

  (defpact transfer-crosschain:string
    ( sender:string
      receiver:string
      receiver-guard:guard
      target-chain:string
      amount:decimal )
    (step (let ((_ (enforce false "todo"))) "")))

  (defun get-balance:decimal (account:string)
    (reputation.get-balance (reputation.get-snap SELF GENESIS) RESOURCE account))

  (defun details:object{fungible-v2.account-details} (account:string)
    { 'account: account
    , 'balance: (get-balance account)
    , 'guard: (at 'guard (coin.details account)) })

  (defun precision:integer () 12)

  (defun enforce-unit:bool (amount:decimal)
    (enforce-precision amount (precision)))

  (defun create-account:string (account:string guard:guard)
    (coin.create-account account guard))

  (defun rotate:string (account:string new-guard:guard)
    (coin.rotate account new-guard))


)


(if (contains "upgrade" (read-msg))
  []
  [ (create-table documents)
    (create-table revisions)
    (initialize) ])
