(namespace (read-msg 'ns))
(define-keyset 'bunny)

(module flow 'bunny
  @doc "utility for implicit computation of proportional areas on a curve. \
      \ assumes constant weights in-between updates. \
      \ orphan flow is redistributed to first outflow with non-null weight"
  @model
  [ ; weight eq
    (property
      (=
        (column-delta inflows 'total-weight)
        (column-delta outflows 'weight)))

    ; flow eq
    (property
      (=
        (column-delta inflows 'total-flow)
        (column-delta outflows 'flow)))

    ; derivatives
    (property (>= (column-delta inflows 'source) 0.0))
    (property (>= (column-delta inflows 'total-flow) 0.0))
    (property (>= (column-delta inflows 'accumulator) 0.0))
    (property (>= (column-delta outflows 'flow) 0.0))
    (property (>= (column-delta outflows 'last-accumulator) 0.0)) ]

  (use util)

  (defconst PRECISION 32)

  (defschema inflow
    @model
      [ (invariant (>= source 0.0))
        (invariant (>= total-weight 0.0))
        (invariant (>= total-flow 0.0))
        (invariant (<= total-flow source))
        (invariant (>= accumulator 0.0)) ]
    domain:module{domain-v0}
    source:decimal
    total-weight:decimal
    total-flow:decimal
    accumulator:decimal )

  (defschema outflow
    @model
      [ (invariant (>= flow 0.0))
        (invariant (>= weight 0.0))
        (invariant (>= last-accumulator 0.0)) ]
    weight:decimal
    flow:decimal
    last-accumulator:decimal
    last-update:time )

  (defschema area
    @model
      [ (invariant (>= flow-delta 0.0))
        (invariant (>= weight-delta 0.0)) ]
    inflow:object{inflow}
    outflow:object{outflow}
    flow-delta:decimal
    weight-delta:decimal
    interval:[time] )

  (deftable inflows:{inflow})
  (deftable outflows:{outflow})

  (defun create-inflow (inflow:string domain:module{domain-v0})
    (enforce-domain domain flow)
    (insert inflows inflow
      { 'domain: domain
      , 'source: 0.0
      , 'total-weight: 0.0
      , 'total-flow: 0.0
      , 'accumulator: 0.0 }))

  (defun create-outflow (inflow:string outflow:string)
    (with-read inflows inflow { 'domain := domain, 'accumulator := acc }
      (enforce-domain domain flow)
      (insert outflows (collection inflow outflow)
        { 'weight: 0.0
        , 'flow: 0.0
        , 'last-accumulator: acc
        , 'last-update: (now) })))

  (defun compute-inflow:object{inflow} (inflow:object{inflow} new-source:decimal)
    @model [ (property (>= new-source 0.0)) ]
    (enforce (>= new-source 0.0) "invalid negative source")
    (let*
      ( (total-weight (at 'total-weight inflow))
        (orphaned (= 0.0 total-weight)) )
      (if orphaned
        inflow
        (let*
          ( (source (at 'source inflow))
            (accumulator (at 'accumulator inflow))
            (source-delta (- new-source source))
            (accumulator-delta (floor (/ source-delta total-weight) PRECISION)) )
          (enforce (>= new-source source) "expects `new-source ≥ source`")
          (enforce (>= total-weight 0.0) "error: unexpected read")
          (enforce (>= accumulator-delta 0.0) "error: unexpected value")
          (+
            { 'source: new-source
            , 'accumulator: (+ accumulator accumulator-delta) }
            inflow)))))

  (defun compute-outflow:object{area}
    ( inflow:object{inflow}
      outflow:object{outflow}
      new-weight:decimal )
    @model [ (property (>= new-weight 0.0)) ]
    (enforce (>= new-weight 0.0) "invalid negative weight")
    (let*
      ( (t (now))
        (weight (at 'weight outflow))
        (new-accumulator (at 'accumulator inflow))
        (accumulator-diff (- new-accumulator (at 'last-accumulator outflow)))
        (flow-delta (floor (* weight accumulator-diff) PRECISION))
        (weight-delta (- new-weight weight))
        (new-total-flow (+ flow-delta (at 'total-flow inflow)))
        (new-total-weight (+ weight-delta (at 'total-weight inflow))) )
    (enforce (>= flow-delta 0.0) "error: unexpected value")
    (enforce (>= accumulator-diff 0.0) "error: unexpected value")
    (enforce
      (and
        (>= new-total-flow 0.0)
        (<= new-total-flow (at 'source inflow)))
      "error: unexpected value")
    (enforce (>= new-total-flow (at 'total-flow inflow)) "baka solver")
    (enforce (>= new-total-weight 0.0) "error: unexpected value")
    { 'inflow:
      (+
        { 'total-flow: new-total-flow
        , 'total-weight: new-total-weight }
        inflow)
    , 'outflow:
      { 'weight: new-weight
      , 'flow: (+ (at 'flow outflow) flow-delta)
      , 'last-accumulator: new-accumulator
      , 'last-update: t }
    , 'flow-delta: flow-delta
    , 'weight-delta: weight-delta
    , 'interval: [ (at 'last-update outflow) t ] }))

  (defun update-flow:object{area}
    ( inflow:string
      new-source:decimal
      outflow:string
      new-weight:decimal )
    @model
      [ (property (>= new-source 0.0))
        (property (>= new-weight 0.0)) ]
    (enforce-precision new-source PRECISION)
    (enforce-precision new-weight PRECISION)
    (let*
      ( (outflow-key (collection inflow outflow))
        (outflow-row (read outflows outflow-key))
        (new-inflow (compute-inflow (get-inflow inflow) new-source))
        (area (compute-outflow new-inflow outflow-row new-weight)) )
      (enforce-domain (at 'domain (at 'inflow area)) flow)
      (update inflows inflow (drop [ 'domain ] (at 'inflow area)))
      (update outflows outflow-key (at 'outflow area))
      area))

  (defun get-inflow:object{inflow} (inflow:string)
    (read inflows inflow))

  (defun get-outflow:object{outflow} (inflow:string outflow:string)
    (read outflows (collection inflow outflow)))


  (implements domain-v0)

  (defcap OPERATE:bool (target:module{domain-v0})
    (enforce false "unsupported"))

)


(if (contains "upgrade" (read-msg))
  []
  [ (create-table inflows)
    (create-table outflows) ])
