(namespace (read-msg 'ns))
(define-keyset 'bunny)

(module elastic 'bunny
  @doc "w(x) = ( ∑ x_{i}^{p} )^{q} \
      \ where w is the weight of group, x_i the shares on each market, \
      \ and (p, q) are cost parameters"

  (defschema parameters
    p:decimal
    q:decimal

    ; precision does not affect shares (nor p-sum) in order to allow caller using
    ; higher precision on those such that every weights can be represented (and solved)
    ; in this respect, extra required precision depends on expected scale of derivatives
    ; (pact's pow function seems limited to ~51 decimals)
    precision:integer )

  (defschema vars
    p:decimal
    q:decimal
    p-sum:decimal
    precision:integer )

  (defun initial:object{vars} (params:object{parameters})
    (bind params { 'p := p, 'q := q, 'precision := precision }
      (enforce (>= p 1.0) "expect `p ≥ 1`")
      (enforce (and (> q 0.0) (<= q 1.0)) "expect `q > 0 ∧ q ≤ 1`")
      (enforce
        (>= (* p q) 1.0)
        "expect `p * q ≥ 1`, second-order partial derivatives must be positive")
      (enforce (>= precision 0) "expect `precision ≥ 0`")
      { 'p: p, 'q: q, 'p-sum: 0.0, 'precision: precision }))

  (defun weight:decimal (vars:object{vars})
    (bind vars { 'q := q, 'p-sum := p-sum, 'precision := precision }
      (round (^ p-sum q) precision)))

  (defun swap:object{vars} (vars:object{vars} x0:decimal x1:decimal)
    (enforce (and (>= x0 0.0) (>= x1 0.0)) "error: illegal arguments")
    (bind vars { 'p := p, 'q := q, 'p-sum := p-sum, 'precision := precision }
      (let ((new-p-sum (+ p-sum (- (^ x1 p) (^ x0 p)))))
        ; TODO: handle precision errors (if any), emptying pool may fail
        (enforce (>= new-p-sum 0.0) "error: illegal value")
        { 'p: p, 'q: q, 'p-sum: new-p-sum, 'precision: precision })))

  (defun solve:decimal (vars:object{vars} shares:decimal dw:decimal)
    (bind vars { 'p := p, 'q := q, 'p-sum := p-sum, 'precision := precision }
      (util.enforce-precision dw precision)
      (let*
        ( (y (- p-sum (^ shares p)))
          (new-weight (+ (weight vars) dw))
          (new-shares (^ (- (^ new-weight (/ 1.0 q)) y) (/ 1.0 p))) )
        (enforce (>= new-weight 0.0) "input weight delta below limit")
        (enforce (>= new-shares 0.0) "unsolvable")
        (- new-shares shares))))

  (defun derivative:decimal (vars:object{vars} shares:decimal)
    @doc "determines marginal price of shares"
    (bind vars { 'p := p, 'q := q, 'p-sum := p-sum }
      (* (* p q) (* (^ shares (- p 1.0)) (^ p-sum (- q 1.0))))))
)


(module poll 'bunny

  (use util)

  (defcap INTERNAL () true)

  (defcap SWAP (group:string market:string actor:string w:decimal s:decimal t:time)
    @event
    true)

  (defschema group
    @model [ (invariant (>= size 0)) ]
    domain:module{domain-v0}
    vars:object{vars}
    size:integer
    t:time )

  (defschema market
    @model [ (invariant (>= shares 0.0)) ]
    shares:decimal
    t:time )

  (defschema position
    @model [ (invariant (>= holdings 0.0)) ]
    holdings:decimal )

  (defschema swap
    group:object{group}
    market:object{market}
    weights:[decimal]
    shares:[decimal]
    holdings:[decimal]
    t:time )

  (deftable groups:{group})
  (deftable markets:{market})
  (deftable positions:{position})

  (defun create-group
    ( group:string
      parameters:object{parameters}
      domain:module{domain-v0} )
    (enforce-domain domain poll)
    (insert groups group
      { 'domain: domain
      , 'vars: (elastic.initial parameters)
      , 'size: 0
      , 't: (now) }))

  (defun create-market (group:string market:string)
    (with-read groups group { 'domain := domain, 'size := size }
      (enforce-domain domain poll)
      (update groups group { 'size: (+ 1 size) } ))
    (insert markets (collection group market)
      { 'shares: 0.0
      , 't: (now) }))

  (defun update-position:[decimal] (key:string ds:decimal)
    (require-capability (INTERNAL))
    (enforce (!= ds 0.0) "null shares delta")
    (with-default-read positions key { 'holdings: 0.0 } { 'holdings := h0 }
      (let ((h1 (+ h0 ds)))
        (enforce (>= h1 0.0) "insufficient shareholdings")
        (write positions key { 'holdings: h1 })
        [ h0 h1 ])))

  (defun compute-swap:object{swap}
    ( group:object{group}
      market:object{market}
      holdings:[decimal]
      ds:decimal )
    (enforce (!= ds 0.0) "null shares delta")
    (enforce (= (length holdings) 2) "error: unexpected argument length")
    (let*
      ( (x0 (at 'shares market))
        (x1 (+ x0 ds))
        (vars (at 'vars group))
        (new-vars (elastic.swap vars x0 x1))
        (new-group (+ { 'vars: new-vars } group))
        (w0 (elastic.weight vars))
        (w1 (elastic.weight new-vars)) )
      (enforce (>= (at 1 holdings) 0.0) "insufficient shares")
      (enforce (>= x1 0.0) "error: insufficient shares in pool")
      (enforce (and (>= w0 0.0) (>= w1 0.0)) "error: unexpected value")
      (enforce (= (signum (- w1 w0)) (signum ds)) "error: unexpected sign")
      { 'group: new-group
      , 'market: (+ { 'shares: x1 } market)
      , 'weights: [ w0 w1 ]
      , 'shares: [ x0 x1 ]
      , 'holdings: holdings
      , 't: (now) }))

  (defun update-market:object{swap}
    ( group:string
      market:string
      actor:string
      ds:decimal )
    @doc "domain must manage precision, actor guards and resource bonds"
    (let*
      ( (market-key (collection group market))
        (position-key (collection market-key actor))
        (group-row (read groups group))
        (market-row (read markets market-key))
        (domain (at 'domain group-row))
        (holdings (with-capability (INTERNAL) (update-position position-key ds)))
        (swap (compute-swap group-row market-row holdings ds)) )
      (enforce-domain domain poll)
      (update groups group (take [ 'vars ] (at 'group swap)))
      (update markets market-key (take [ 'shares ] (at 'market swap)))
      (emit-event (SWAP group market actor
        (at 1 (at 'weights swap)) (at 1 (at 'shares swap)) (at 't swap)))
      swap))


  (defun get-group:object{group} (group:string)
    (read groups group))

  (defun get-market:object{market} (group:string market:string)
    (read markets (collection group market)))

  (defun get-position:decimal (group:string market:string actor:string)
    (with-default-read positions (collection (collection group market) actor)
      { 'holdings: 0.0 }
      { 'holdings := h }
      h))

  (defun get-weight:decimal (group:string)
    (with-read groups group { 'vars := vars }
      (elastic.weight vars)))

  (defun solve:decimal (group:string market:string dw:decimal)
    (with-read groups group { 'vars := vars }
      (with-read markets (collection group market) { 'shares := shares }
        (elastic.solve vars shares dw))))

  (defun view-swap:object{swap}
    ( group:string
      market:string
      actor:string
      ds:decimal )
    (let*
      ( (h0 (get-position group market actor))
        (holdings [ h0 (+ h0 ds) ]) )
      (compute-swap (get-group group) (get-market group market) holdings ds)))


  (implements domain-v0)

  (defcap OPERATE:bool (target:module{domain-v0})
    (enforce false "unsupported"))

)

(if (contains "upgrade" (read-msg))
  []
  [ (create-table groups)
    (create-table markets)
    (create-table positions) ])
