(namespace (read-msg 'ns))
(define-keyset 'bunny)

(interface reputation-policy-v0

  (defschema outcome
    @model
      [ (invariant (>= shares 0.0))
        (invariant (>= lifetime 0.0)) ]
    shares:decimal
    lifetime:decimal )

  (defun parameters:object{elastic.parameters} ()
    @doc "const")

  (defun learning-curve:decimal (t:decimal)
    @doc "internal inflation. df/dt ≥ 0"
    @model
     [ (property (>= t 0.0))
       (property (>= result 0.0)) ])

 (defun seeding-curve:decimal (t:decimal)
   @doc "external inflation. df/dt ≥ 0"
   @model
    [ (property (>= t 0.0))
      (property (> result 0.0)) ])

 (defun density:decimal (t:decimal)
   @doc "deflates inactive supply. df/dt ≤ 0"
   @model
    [ (property (>= t 0.0))
      (property (and (>= result 0.0) (<= result 1.0))) ])

  (defun reflow:decimal (t:decimal outcome:object{outcome})
    @doc "determines inflation ratio (from learning curve)"
    @model
      [ (property (>= t 0.0))
        (property (>= result 0.0)) ])

)



(module reputation 'bunny
  @doc "self-referential reputation systems"

  (use util)
  (use poll [ swap create-group create-market update-market ])
  (use flow [ area inflow create-inflow create-outflow update-flow get-inflow ])

  (defcap INTERNAL () true)

  (defconst ACTOR_NULL { 'fragments: 0.0 })
  (defconst SHARES_EXTRA_PRECISION 6
    @doc "ensure most amounts can be represented by shares solver (heuristic)")

  (defschema resource
    domain:module{domain-v0}
    policy:module{reputation-policy-v0}
    float:decimal
    seeded:decimal
    genesis:time )

  (defschema actor
    ; fragments are the internal representation of inactive reputation tokens
    ; both have the same precision, since convertion factor is a whole
    ; number defining the number of fragments that equals 1 token
    fragments:decimal )

  (defschema submission
    origin:string
    frozen-until:time
    staked-until:time )

  (defschema snap
    @doc "tokenomics snapshot"
    learn:decimal
    seed:decimal
    source:decimal
    base:decimal
    precision:integer
    t:decimal )

  (defschema mark
    @doc "vote result"

    resource:string
    submission:object{submission}
    swap:object{swap}
    area:object{area}
    snap:object{snap}

    ; cost relative to actor (positive -> debit, negative -> credit)
    cost:decimal )

  (deftable resources:{resource})
  (deftable actors:{actor})
  (deftable submissions:{submission})

  (defun create-resource
    ( resource:string
      policy:module{reputation-policy-v0}
      domain:module{domain-v0} )
    (enforce-domain domain reputation)
    (with-capability (OPERATE poll)
      (create-group resource (try false (policy::parameters)) reputation))
    (with-capability (OPERATE flow)
      (create-inflow resource reputation))
    (insert resources resource
      { 'domain: domain
      , 'policy: policy
      , 'float: 0.0
      , 'seeded: 0.0
      , 'genesis: (now) }))

  (defun create-submission
    ( resource:string
      submission:string
      origin:string
      stake:object{stake} )
    (validate-stake stake)
    (with-read resources resource { 'domain := domain }
      (enforce-domain domain reputation))
    (with-capability (OPERATE poll)
      (create-market resource submission))
    (with-capability (OPERATE flow)
      (create-outflow resource submission))
    (insert submissions (collection resource submission)
      { 'origin: origin
      , 'frozen-until: (add-time (now) (days -7))
      , 'staked-until: (add-time (now) (at 'duration stake)) })
    (if (> (at 'amount stake) 0.0)
      [ (vote resource submission origin
          (solve resource submission (at 'amount stake))) ]
      []))

  (defun update-actor (resource:string actor:string delta:decimal)
    (require-capability (INTERNAL))
    (let ((key (collection resource actor)))
      (with-default-read actors key ACTOR_NULL { 'fragments := fragments }
        (let ((newfrags (+ fragments delta)))
          (enforce (>= newfrags 0.0) "insufficient balance")
          (write actors key { 'fragments: newfrags })))))

  (defun update-pool
    ( resource:string
      actor:string
      snap:object{snap}
      amount:decimal )
    (require-capability (INTERNAL))
    (with-read resources resource { 'float := float }
      (let*
        ( (delta (from-denomination snap amount))
          (new-float (+ float delta)) )
        (update-actor resource actor delta)
        (enforce (>= new-float 0.0) "error: unexpected state")
        (enforce
          (<= new-float (from-denomination snap (at 'source snap)))
          "error: unexpected state")
        (update resources resource { 'float: new-float }))))

  (defun seed (resource:string actor:string amount:decimal)
    (enforce (> amount 0.0) "error: invalid amount")
    (with-read resources resource
      { 'domain := domain
      , 'policy := policy
      , 'seeded := s0
      , 'genesis := genesis }
      (enforce-domain domain reputation)
      (let*
        ( (snap (get-snap policy genesis))
          (max-seed (at 'seed snap))
          (s1 (+ s0 amount)) )
        (enforce (<= s1 max-seed) "error: exceeded seed limit")
        (update resources resource { 'seeded: s1 })
        (with-capability (INTERNAL)
          (update-pool resource actor snap amount)))))

  (defun transfer:string (resource:string from:string to:string amount:decimal)
    ; will ensure both balances are changed by the same amount
    (enforce (> amount 0.0) "invalid transfer amount")
    (with-read resources resource
      { 'domain := domain
      , 'policy := policy
      , 'genesis := genesis }
      (enforce-domain domain reputation)
      (let*
        ( (snap (get-snap policy genesis))
          (fragments (from-denomination snap amount)) )
        (with-capability (INTERNAL)
          (update-actor resource from (- fragments))
          (update-actor resource to fragments)))))

  (defun vote:decimal
    ( resource:string
      submission:string
      actor:string
      ds:decimal )
    (enforce (!= ds 0.0) "null shares delta")
    (with-read resources resource
      { 'domain := domain, 'policy := poli, 'genesis := genesis }
      (let*
        ( (submission-row (get-submission resource submission))
          (policy:module{reputation-policy-v0} poli)
          (snap (get-snap policy genesis))
          (swap
            (with-capability (OPERATE poll)
              (update-market resource submission actor ds)))
          (precision (at 'precision snap))
          (cost (- (at 1 (at 'weights swap)) (at 0 (at 'weights swap))))
          (flow-weight
            (round
              (try false (policy::reflow (at 't snap) (get-outcome swap)))
              precision))
          (area
            (with-capability (OPERATE flow)
              (update-flow resource (at 'learn snap) submission flow-weight))) )

        (enforce-domain domain reputation)
        (enforce-precision ds (+ SHARES_EXTRA_PRECISION precision))

        ; following enforcement also prevents null cost (from rounding errors)
        (enforce (= (signum ds) (signum cost)) "error: unexpected signs")

        (with-capability (INTERNAL)
          (update-pool resource actor snap (- cost)))

        (bind submission-row
          { 'origin := origin
          , 'frozen-until := frozen-until
          , 'staked-until := staked-until }
          (enforce (>= (now) frozen-until) "submission is frozen")
          (enforce
            (or
              (!= actor origin)
              (or
                (>= (now) staked-until)
                (= (signum ds) 1.0)))
            "staked position can not be sold yet")
          (with-capability (INTERNAL)
            (update-pool resource origin snap
              (floor (at 'flow-delta area) precision))))

        { 'resource: resource
        , 'submission: submission-row
        , 'swap: swap
        , 'area: area
        , 'snap: snap
        , 'cost: cost })))

  (defun solve:decimal (resource:string submission:string cost:decimal)
    (with-read resources resource { 'policy := policy }
      (let ((precision (get-precision policy)))
        (enforce-precision cost precision)
        (round
          (poll.solve resource submission cost)
          (+ SHARES_EXTRA_PRECISION precision)))))


  (defun rebase:decimal (density:decimal precision:integer)
    (enforce (> density 0.0) "error: unexpected value")
    (enforce (<= density 1.0) "error: unsupported")
    (round (* (^ 10.0 precision) (/ 1.0 density)) 0))

  (defun to-denomination:decimal (snap:object{snap} fragments:decimal)
    (enforce (>= fragments 0.0) "error: invalid value")
    (bind snap { 'precision := precision, 'base := base }
      (enforce-precision fragments precision)
      (floor (/ fragments base) precision)))

  (defun from-denomination:decimal (snap:object{snap} amount:decimal)
    ; expects base to be a whole number
    (bind snap { 'precision := precision, 'base := base }
      (enforce-precision amount precision)
      (* amount base)))

  (defun get-precision:integer (policy:module{reputation-policy-v0})
    (at 'precision (try false (policy::parameters))))

  (defun compute-snap:object{snap} (policy:module{reputation-policy-v0} t:decimal)
    (enforce (>= t 0.0) "error: unexpected argument")
    (let*
      ( (precision (get-precision policy))
        (learn (floor (try false (policy::learning-curve t)) precision))
        (seed (floor (try false (policy::seeding-curve t)) precision))
        (base (rebase (try false (policy::density t)) precision)) )
    { 'learn: learn
    , 'seed: seed
    , 'source: (+ learn seed)
    , 'base: base
    , 'precision: precision
    , 't: t }))

  (defun get-outcome:object{outcome} (swap:object{swap})
    { 'shares: (at 1 (at 'shares swap))
    , 'lifetime: (elapsed (at 't (at 'market swap))) })

  (defun get-snap:object{snap} (policy:module{reputation-policy-v0} genesis:time)
    (compute-snap policy (elapsed genesis)))

  (defun get-resource:object{resource} (resource:string)
    (read resources resource))

  (defun get-submission:object{submission} (resource:string submission:string)
    (read submissions (collection resource submission)))

  (defun get-actor:object{actor} (resource:string actor:string)
    (read actors (collection resource actor)))

  (defun get-balance:decimal (snap:object{snap} resource:string actor:string)
    (with-default-read actors (collection resource actor)
      ACTOR_NULL { 'fragments := fragments }
      (to-denomination snap fragments)))

  (defun get-float:decimal (resource:string)
    (with-read resources resource
      { 'float := float, 'policy := policy, 'genesis := genesis }
      (to-denomination (get-snap policy genesis) float)))

  (defun get-supply:decimal (resource:string)
    @doc "this function ignores future potential mints (unrealized flow or seed) \
        \ actual supply will be less due to unreachable fragments on individual accounts"
    (+ (poll.get-weight resource) (get-float resource)))


  (implements domain-v0)

  (defcap OPERATE:bool (target:module{domain-v0}) true)

)


(if (contains "upgrade" (read-msg))
  []
  [ (create-table resources)
    (create-table actors)
    (create-table submissions) ])
