(namespace (read-msg 'ns))
(define-keyset 'bunny)

(module util 'bunny

  (defconst SINGLETON "")

  (defconst PI 3.14159265358979323846264338327950288419716939937510582097494459)

  (defconst PHI (/ (+ 1.0 (sqrt 5.0)) 2.0))


  ; ------------------------------------------------------------------------
  ; date

  (defconst EPOCH (time "1970-01-01T00:00:00Z"))

  (defun now:time ()
    (at "block-time" (chain-data)))

  (defun seconds:decimal (x:decimal)
    x)

  (defun months:decimal (x:decimal)
    (* x (days 30.4375)))

  (defun years:decimal (x:decimal)
    (* x (days 365.25)))

  (defun elapsed:decimal (genesis:time)
    @model [ (property (>= result 0.0)) ]
    (let ((t (diff-time (now) genesis)))
      (enforce (>= t 0.0) "unexpected chain or genesis time")
      t))


  ; ------------------------------------------------------------------------
  ; accounting

  (defun enforce-identity:bool (b:bool)
    (enforce b "enforcement failed"))

  (defun always:guard ()
    (create-user-guard (enforce-identity true)))

  (defun never:guard ()
    (create-user-guard (enforce-identity false)))

  (defun enforce-domain:bool (domain:module{domain-v0} target:module{domain-v0})
    (require-capability (domain::OPERATE target)))

  (defun collection:string (group:string item:string)
    @doc "simple key generation. resulting identifier must be unwrapped from right to left"
    @model [ (property (!= result "")) ]
    (enforce (not (contains ";" item)) "prohibited character: ;")
    (format "{};{}" [ group item ]))

  (defun linear-transfer-manager:decimal (managed:decimal requested:decimal)
    (let ((newbal (- managed requested)))
      (enforce (>= newbal 0.0) "amount transfer exceeded")
      newbal))

  (defun enforce-precision:bool (amount:decimal precision:integer)
    (enforce
      (= (floor amount precision) amount)
      (format "amount violates minimum precision: {}" [ amount ])))

  (defschema stake
    amount:decimal
    duration:decimal )

  (defun stake-null:object{stake} () { 'amount: 0.0, 'duration: 0.0 })

  (defun validate-stake:bool (stake:object{stake})
    (enforce (>= (at 'amount stake) 0.0) "invalid stake amount")
    (enforce (>= (at 'duration stake) 0.0) "invalid stake duration"))


  ; ------------------------------------------------------------------------
  ; math

  (defun signum:decimal (x:decimal)
    (if (= x 0.0) 0.0 (if (> x 0.0) 1.0 -1.0)))

  (defun min:decimal (x:decimal y:decimal)
    @model [ (property (and (<= result x) (<= result y))) ]
    (if (< x y) x y))

  (defun max:decimal (x:decimal y:decimal)
    @model [ (property (and (>= result x) (>= result y))) ]
    (if (< x y) y x))

  (defun clamp:decimal (lo:decimal hi:decimal x:decimal)
    @model
      [ (property (<= lo hi))
        (property (and (>= result lo) (<= result hi))) ]
    (enforce (<= lo hi) "invalid arguments")
    (max lo (min hi x)))

  (defun between:bool (lo:integer hi:integer x:integer)
    @model [ (property (<= lo hi)) ]
    (enforce (<= lo hi) "invalid arguments")
    (and (>= x lo) (<= x hi)))

  (defun every:bool (bs:[bool])
    (fold (and) true bs))

  (defun any:bool (bs:[bool])
    (fold (or) false bs))

  (defun sech:decimal (x:decimal)
    (/ 2.0 (+ (exp x) (exp (- x)))))

  (defun tanh:decimal (x:decimal)
    (/ (- (exp (* 2.0 x)) 1.0) (+ (exp (* 2.0 x)) 1.0)))

  (defun logistic-pdf:decimal (x:decimal mu:decimal s:decimal)
    (enforce (> s 0.0) "invalid argument")
    (* (/ 1.0 (* 4.0 s)) (^ (sech (/ (- x mu) (* 2.0 s))) 2.0)))

  (defun logistic-cdf:decimal (x:decimal mu:decimal s:decimal)
    (enforce (> s 0.0) "invalid argument")
    (+ (/ 1.0 2.0) (* (/ 1.0 2.0) (tanh (/ (- x mu) (* 2.0 s))))))



)
