import { createContext, h } from "preact"
import { useContext, useState, useEffect } from "preact/hooks"
import Pact from "pact-lang-api"
import Transact from "../component/layout/transact"
import { toNative, fromNative, alignIndent, toPact } from "../util/pact"
import { create } from "ipfs-http-client"

const Env = {
  host: "http://localhost:8080",
  explorer: "",
  namespace: "free",
  cacheTTL: 86400, // seconds
  ipfs: {
    gateway: "https://infura-ipfs.io",
    api: "https://ipfs.infura.io:5001"
  },
  defs: {
    chainId: "0",
    gasLimit: 4000,
    gasPrice: 0.0000001,
    ttl: 600,
    networkId: undefined
  }
}

class API {

  static mkCmd = (pactCode, envData = {}, caps = []) => ({
    pactCode: alignIndent(pactCode),
    envData,
    caps,
    ...Env.defs
  })

  static local = async (cmd, data = undefined, caps = undefined) => {
    if (typeof cmd == "string")
      cmd = API.mkCmd(cmd, data, caps)
    let response = await Pact.fetch.local(cmd, Env.host)
    if (response.result.status != "success")
      throw response
    return toNative(response.result.data)
  }

  static sign = (cmd) =>
    Pact.wallet.sign(cmd, Env.host)

  static send = (signedCmd) =>
    Pact.wallet.sendSigned(signedCmd, Env.host)

  static query = {

    state: () => API.local(`
      (namespace "${Env.namespace}")
      (use awa)
      (use awa-persistent-v0)
      { 'constant:
        { 'resource: RESOURCE
        , 'precision: (precision)
        , 'maxNameLength: MAX_NAME_LENGTH
        , 'maxDataLength: MAX_DATA_LENGTH
        , 'ratings: RATINGS }
      , 'accounts:
        { 'treasury: (details TREASURY)
        , 'burns: (details BURNS) }
      , 'snap: (reputation.get-snap awa GENESIS)
      , 'resource: (reputation.get-resource RESOURCE)
      , 'totalSupply: (total-supply)
      , 'stake: (get-stake)
      , 'seedPrice: (get-seed-price) }
    `),

    document: (document) => API.local(`
      (${Env.namespace}.awa.get-document (read-msg 'document))
    `, { document }),

    validateDocument: (document) => API.local(`
      (namespace "${Env.namespace}")
      (let ((doc (read-msg 'document)))
        [ (try true (< 0 (at 'revisions (awa.get-document doc))))
          (try false (awa.validate-name doc)) ])
    `, { document }),

    account: (account) => API.local(`
      (let ((account (read-msg 'account)))
        { 'coin: (coin.details account)
        , 'awa: (${Env.namespace}.awa.details account)})
    `, { account })
  }

  static capability = {

    document: (name, account) => Pact.lang.mkCap(
      "create document",
      "initialize a document with a name and its first revision",
      Env.namespace + ".awa.DOCUMENT",
      [ name, account ]
    ),

    coin: {

      transfer: (from, to, amount) => Pact.lang.mkCap(
        "coin transfer",
        "transfer designated amount of coin from sender to receiver",
        "coin.TRANSFER",
        [ from, to, fromNative(amount) ]
      )

    }

  }

  static command = {

    createDocument: (name, account, data) => API.mkCmd(`
      (${Env.namespace}.awa.create-document ${toPact(name)} ${toPact(account)} (read-msg 'data))
    `, { data }, [ API.capability.document(name, account) ]),

    createDocumentWithDeposit: (name, account, data, treasury, deposit) => API.mkCmd(`
      (namespace "${Env.namespace}")
      (awa.deposit ${toPact(account)} ${toPact(deposit)})
      (awa.create-document ${toPact(name)} ${toPact(account)} (read-msg 'data))
    `, { data }, [
      API.capability.coin.transfer(account, treasury, deposit),
      API.capability.document(name, account)
    ])

  }
}


const Rema = createContext()

const RemaProvider = ({ children }) => {
  const [ state, setState ] = useState(null)
  const [ ipfs, setIpfs ] = useState(null)
  const [ history, setHistory ] = useState({ l: [] })

  // TODO default user

  useEffect(() => {
    Promise.all([
      API.query.state().then(setState)
    ])

    setIpfs(create(new URL(Env.ipfs.api)))
  }, [])

  const commands = {
    get history() {
      return history.l
    },
    dispatch: (o) => {
      const ref = {
        ...o,
        status: "pending",
        onCancel: () => { ref.status = "canceled"; commands.update() },
        onClose: () => {
          if (o.onClose) o.onClose()
          ref.status = "closed"
          commands.update()
        }
      }
      commands.update(l => l.concat(ref))
    },
    update: (f = x => x) => setHistory(o => ({ l: f(o.l) }))
  }

  const value = {
    ipfs,
    state,
    commands
  }
  return <Rema.Provider value={value}>{children}</Rema.Provider>
}

const CommandQueue = () => {
  const rema = useContext(Rema)

  const obj = rema.commands.history.find(o => o.status == "pending")
  return obj ? <Transact {...obj} /> : null
}

export { Env, API, Rema, RemaProvider, CommandQueue }
