import { h, Fragment } from "preact"
import { useContext, useState, useRef, useEffect } from "preact/hooks"
import Header from "../component/layout/header"
import Footer from "../component/layout/footer"
import { Button, Label } from "../component/shared/button"
import Input from "../component/shared/input"
import { Env, API, Rema } from "../context/rema"
import { CID } from "ipfs-http-client"
import BigDecimal from "../util/bigdecimal"
import { Arrow } from "../component/shared/icon"

// TODO rename to "editor" and use this page for both new and edit

textareaPlaceholder = `text content

- everything will be saved forever
- only content-addressed files will be loaded
- submission should be in public domain or equivalent
- be respectful ♡

markdown
# title 1
## title 2
### title 3
link: [text](document name) or [text](url)
image: ![alt text](ipfs cid)`

const documentErrors = [ "document already exists", "invalid name" ]

const computeStakeCost = (state, account) => {
  let seedPrice = state.seedPrice, amount = state.stake.amount, balance = account.awa.balance
  return [
    amount.subtract(balance).max(0).multiply(seedPrice),
    amount.min(balance)
  ]
}

// TODO denomination object or component?
const showDenomination = (amount, ccy) =>
  <><strong>{amount.toFixed(1, ",")}</strong> <span class="ccy">{ccy}</span></>

const mkStakeTable = (state, account) => {
  const [ kda, awa ] = computeStakeCost(state, account)
  return <div class="rows hor-margin">
    <div>
      <Label><span>stake duration: <strong>{state.stake.duration.divide(86400).toFixed(1, ",")} days</strong></span></Label>
      <Label><span> {showDenomination(new BigDecimal(1), "owo")} = {showDenomination(state.seedPrice, "kda")}</span></Label>
      { kda > 0.0
        ? <Label><span>available seed: {showDenomination(state.snap.seed.subtract(state.resource.seeded), "owo")}</span></Label>
        : null }
    </div>
    <table style="align-self: flex-start;">
      <tr>
        <th>ccy</th>
        <th>balance</th>
        <th class="warn color">change</th>
      </tr>
      <tr>
        <td class="ccy">kda</td>
        <td class="numeric">{account.coin.balance.toFixed(2, ",")}</td>
        <td class={"numeric color " + (kda.n > account.coin.balance.n ? "important error" : "warn")}>{kda.multiply(-1).toFixed(2, ",")}</td>
      </tr>
      <tr>
        <td class="ccy">owo</td>
        <td class="numeric">{account.awa.balance.toFixed(2, ",")}</td>
        <td class="numeric warn color">{awa.multiply(-1).toFixed(2, ",")}</td>
      </tr>
    </table>
  </div>
}

export default ({ name = "" }) => {
  const rema = useContext(Rema)
  const [ account, setAccount ] = useState(null)
  const [ dataCid, setDataCid ] = useState(null)
  const textarea = useRef(null)

  const uploadTextarea = (e) => {
    e.preventDefault()
    rema.ipfs.add(textarea.current.value).then(({cid}) => setDataCid(cid.toString()))
  }

  const submit = (e) => {
    e.preventDefault()
    const inputs = [ "name", "account", "data" ].map(k => e.target.elements[k])
    if (!account || inputs.some(el => el.classList.contains("invalid"))) return;
    let [ kda, awa ] = computeSeedTarget(rema.state, account)
    rema.commands.dispatch({
      cmd: (kda.n > 0 ? API.command.createDocumentWithDeposit : API.command.createDocument)(
        ...inputs.map(el => el.value),
        rema.state.account.treasury.account,
        kda
      )
    })
  }

  return <>
    <Header path={[ "special", "new" ]}/>
    <section>
      <main>
        <article>
          <form onSubmit={submit} class="rows x2">
            <h1>New document</h1>
            <fieldset class="rows">
              <legend>document name</legend>
              <Input
                placeholder=" "
                name="name"
                value={ name }
                validate={n => API.query.validateDocument(n)
                  .then(es => es
                    .map((e, i) => !e ? documentErrors[i] : null)
                    .filter(e => e)
                    .join(",")) }
                required />
            </fieldset>
            <fieldset class="rows">
              <legend>coin account</legend>
              <Input
                placeholder=" "
                name="account"
                validate={(id) => {
                  setAccount(null);
                  return API.query.account(id)
                    .then(o => { setAccount(o); return "" })
                    .catch(_ => "could not retrieve account") }}
                required />
              <div class="columns hor-margin justify-start">
                <Label align="lowercase" class="important">testnet chain 1</Label>
                <Button align="lowercase" class="dim" href="https://faucet.testnet.chainweb.com/">faucet</Button>
              </div>
              { rema.state && account ? mkStakeTable(rema.state, account) : null}
            </fieldset>
            <fieldset class="rows">
              <legend>initial data</legend>
              <Input
                placeholder="ipfs cid"
                validate={ x => {
                  try { CID.parse(x); return Promise.resolve("") }
                  catch(e) { return Promise.resolve("invalid cid") } }}
                name="data"
                value={dataCid ?? ""}
                setValue={setDataCid}
                required />
              <strong class="label dim color">――&nbsp;&nbsp;or&nbsp;&nbsp;――</strong>
              <div class="columns justify-start pack hor-margin">
                <Label align="lowercase" class="dimmer collapse-right shadow">{Env.ipfs.api}</Label>
                <Button align="lowercase" class="dim collapse-left bg shadow" onClick={uploadTextarea}><Arrow angle="90" /> upload</Button>
              </div>
              <textarea placeholder={textareaPlaceholder} class="input" style="min-height: 22rem; resize: vertical;" ref={textarea}></textarea>
            </fieldset>
            <fieldset class="columns justify-center">
              <Button align="lowercase" class="important" href="/" tabindex="-1">cancel</Button>
              <Button type="submit" align="lowercase" class="important secondary invert">next</Button>
            </fieldset>
          </form>
        </article>
      </main>
    </section>
    <Footer />
  </>
}
