import Strings from "./strings"

export default class BigDecimal {
  // always truncates excess precision
  // based on https://stackoverflow.com/questions/16742578/bigdecimal-in-javascript

  static PRECISION = 32
  static SHIFT = BigInt("1" + "0".repeat(BigDecimal.PRECISION))
  static ZERO = new BigDecimal("0.0")

  static from = (bigint) =>
    Object.assign(Object.create(BigDecimal.prototype), { n: bigint })

  constructor(value) {
    if (value instanceof BigDecimal)
      return value
    let rep = typeof value == "number" ? value.toFixed(BigDecimal.PRECISION) : String(value)
    let [int, dec] = rep.split(".").concat("")
    this.n = BigInt(int + dec.padEnd(BigDecimal.PRECISION, "0").slice(0, BigDecimal.PRECISION))
  }

  add(x) {
    return BigDecimal.from(this.n + new BigDecimal(x).n)
  }

  subtract(x) {
    return BigDecimal.from(this.n - new BigDecimal(x).n)
  }

  multiply(x) {
    return BigDecimal.from((this.n * new BigDecimal(x).n) / BigDecimal.SHIFT)
  }

  divide(x) {
    return BigDecimal.from((this.n * BigDecimal.SHIFT) / new BigDecimal(x).n)
  }

  pow(x) {
    return new BigDecimal(Math.pow(this.toNumber(), Number(x)))
  }


  min(x) {
    const a = new BigDecimal(x)
    return a.n < this.n ? a : this
  }

  max(x) {
    const a = new BigDecimal(x)
    return a.n > this.n ? a : this
  }


  fragments() {
    const negative = this.n < 0,
      s = this.n.toString().slice(negative).padStart(BigDecimal.PRECISION + 1, "0")
    return [
      negative ? "-" : "+",
      s.slice(0, - BigDecimal.PRECISION),
      s.slice(- BigDecimal.PRECISION)
    ]
  }

  format({
    signed = false,
    separator = ".",
    grouping = "",
    minimumDecimals = 0,
    maximumDecimals = null,
    trailingZero = false
  } = {}) {
    let [sign, int, dec] = this.fragments()
    sign = signed || sign == "-" ? sign : ""
    dec = trailingZero ? dec : dec.replace(/0*$/, "")

    if (grouping != "")
      int = Strings.groups(int, 3).join(grouping)

    if (minimumDecimals > 0)
      dec = dec.padEnd(minimumDecimals, "0")

    if (maximumDecimals != null)
      dec = dec.slice(0, maximumDecimals)

    return sign + int + (dec.length ? separator + dec : "")
  }

  toFixed(precision, grouping = "") {
    return this.format({ grouping, minimumDecimals: precision, maximumDecimals: precision })
  }

  toString() {
    return this.format()
  }

  toNumber() {
    return parseFloat(this.toString())
  }
}
